# standard libraries
import contextlib
import copy
import logging
import random
import threading
import time
import unittest
import uuid

# third party libraries
import numpy
import scipy
import scipy.fft

# local libraries
from nion.data import Calibration
from nion.data import Core
from nion.data import Image
from nion.swift import Application
from nion.swift import Facade
from nion.swift.model import DataItem
from nion.swift.model import DocumentModel
from nion.swift.model import Graphics
from nion.swift.model import Symbolic
from nion.swift.model import Utility
from nion.swift.test import TestContext
from nion.ui import TestUI
from nion.utils import Geometry


Facade.initialize()


class TestSymbolicClass(unittest.TestCase):

    def setUp(self):
        TestContext.begin_leaks()
        self.app = Application.Application(TestUI.UserInterface(), set_global=True)

    def tearDown(self):
        TestContext.end_leaks(self)

    def test_unary_inversion_returns_inverted_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, -d)

    def test_binary_addition_returns_added_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d1 = numpy.zeros((8, 8), dtype=numpy.uint32)
            d1[:] = random.randint(1, 100)
            data_item1 = DataItem.DataItem(d1)
            d2 = numpy.zeros((8, 8), dtype=numpy.uint32)
            d2[:] = random.randint(1, 100)
            data_item2 = DataItem.DataItem(d2)
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + b.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item1))
            computation.create_input_item("b", Symbolic.make_item(data_item2))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d1 + d2)

    def test_binary_multiplication_with_scalar_returns_multiplied_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation1 = document_model.create_computation(Symbolic.xdata_expression("a.xdata * 5"))
            computation1.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation1)
            data1 = DocumentModel.evaluate_data(computation1).data
            computation2 = document_model.create_computation(Symbolic.xdata_expression("5 * a.xdata"))
            computation2.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation2)
            data2 = DocumentModel.evaluate_data(computation2).data
            assert numpy.array_equal(data1, d * 5)
            assert numpy.array_equal(data2, d * 5)

    def test_subtract_min_returns_subtracted_min(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata - numpy.amin(a.data)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d - numpy.amin(d))

    def test_ability_to_take_slice(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.zeros((4, 8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[:,4,4]"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[:,4,4])

    def test_ability_to_take_slice_on_1d_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(8)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[2:6]"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[2:6])

    def test_slice_with_empty_dimension_produces_error(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.zeros((4, 8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[2:2, :, :]"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            self.assertIsNone(DocumentModel.evaluate_data(computation))
            computation.close()

    def test_ability_to_take_slice_with_ellipses_produces_correct_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.zeros((4, 8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[2, ...]"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[2, ...])

    def test_ability_to_take_slice_with_ellipses_produces_correct_calibration(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.zeros((4, 8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            data_item.set_dimensional_calibrations([Calibration.Calibration(10, 20, "m"), Calibration.Calibration(11, 21, "mm"), Calibration.Calibration(12, 22, "nm")])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[2, ...]"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(len(data_and_metadata.data_shape), len(data_and_metadata.dimensional_calibrations))
            self.assertEqual("mm", data_and_metadata.dimensional_calibrations[0].units)
            self.assertEqual("nm", data_and_metadata.dimensional_calibrations[1].units)

    def test_ability_to_take_slice_with_newaxis(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[numpy.newaxis, ...]"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[numpy.newaxis, ...])

    def test_ability_to_take_1d_slice_with_newaxis(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.zeros((8,), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[..., numpy.newaxis]"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[..., numpy.newaxis])

    def test_slice_sum_sums_correct_slices(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(4, 4, 16)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.slice_sum(a.xdata, 4, 6)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.sum(d[..., 1:7], -1))

    def test_reshape_1d_to_2d_produces_correct_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(4)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (2, 2))"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.reshape(d, (2, 2)))
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (4, -1))"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.reshape(d, (4, -1)))
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (-1, 4))"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.reshape(d, (-1, 4)))

    def test_reshape_1d_to_2d_preserves_calibration(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(4)
            data_item = DataItem.DataItem(d)
            data_item.set_dimensional_calibrations([Calibration.Calibration(1.1, 2.1, "m")])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (4, -1))"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual("m", data_and_metadata.dimensional_calibrations[0].units)
            self.assertEqual("", data_and_metadata.dimensional_calibrations[1].units)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (-1, 4))"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual("", data_and_metadata.dimensional_calibrations[0].units)
            self.assertEqual("m", data_and_metadata.dimensional_calibrations[1].units)

    def test_reshape_2d_n_x_1_to_1d_preserves_calibration(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(4, 1)
            data_item = DataItem.DataItem(d)
            data_item.set_dimensional_calibrations([Calibration.Calibration(1.1, 2.1, "m"), Calibration.Calibration()])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (4, ))"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(1, len(data_and_metadata.dimensional_calibrations))
            self.assertEqual("m", data_and_metadata.dimensional_calibrations[0].units)

    def test_reshape_to_match_another_item(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(4)
            d2 = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            data_item2 = DataItem.DataItem(d2)
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, b.xdata.data_shape)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computation.create_input_item("b", Symbolic.make_item(data_item2))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.reshape(d, (2, 2)))

    def test_concatenate_two_images(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(4, 4)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.concatenate((a.xdata[0:2, 0:2], a.xdata[2:4, 2:4]))"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.concatenate((d[0:2, 0:2], d[2:4, 2:4])))

    def test_concatenate_keeps_calibrations_in_non_axis_dimensions(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(4, 4)
            data_item = DataItem.DataItem(d)
            data_item.set_intensity_calibration(Calibration.Calibration(1.0, 2.0, "nm"))
            data_item.set_dimensional_calibrations([Calibration.Calibration(1.1, 2.1, "m"), Calibration.Calibration(1.2, 2.2, "s")])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.concatenate((a.xdata[0:2, 0:2], a.xdata[0:2, 2:4]))"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual("nm", data_and_metadata.intensity_calibration.units)
            self.assertEqual("m", data_and_metadata.dimensional_calibrations[0].units)
            self.assertEqual("", data_and_metadata.dimensional_calibrations[1].units)

    def test_concatenate_along_alternate_axis_images(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(4)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.concatenate((xd.reshape(a.xdata, (1, -1)), xd.reshape(a.xdata, (1, -1))), 0)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.concatenate((numpy.reshape(d, (1, -1)), numpy.reshape(d, (1, -1))), 0))

    def test_concatenate_three_images_along_second_axis(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(4, 4)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.concatenate((a.xdata[0:2, 0:2], a.xdata[1:3, 1:3], a.xdata[2:4, 2:4]), 1)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.concatenate((d[0:2, 0:2], d[1:3, 1:3], d[2:4, 2:4]), 1))

    def test_ability_to_write_read_basic_nodes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.zeros((8, 8), dtype=numpy.uint32)
            src_data[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata / numpy.average(a.data) * 5"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            data_node_dict = computation.write_to_dict()
            data_node_dict["uuid"] = str(uuid.uuid4())
            data_node_dict["variables"][0]["uuid"] = str(uuid.uuid4())
            computation2 = document_model.create_computation()
            computation2.read_from_dict(data_node_dict)
            computation2.needs_update = True
            document_model.append_computation(computation)
            document_model.append_computation(computation2)
            data = DocumentModel.evaluate_data(computation).data
            data2 = DocumentModel.evaluate_data(computation2).data
            assert numpy.array_equal(data, -src_data / numpy.average(src_data) * 5)
            assert numpy.array_equal(data, data2)

    def test_make_operation_works_without_exception_and_produces_correct_data(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata / numpy.average(a.data) * 5"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, -d / numpy.average(d) * 5)

    def test_fft_returns_complex_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.fft(a.xdata)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, scipy.fft.fftshift(scipy.fft.fft2(d) * 1.0 / numpy.sqrt(d.shape[1] * d.shape[0])))

    def test_gaussian_blur_handles_scalar_argument(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, 4.0)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, scipy.ndimage.gaussian_filter(d, sigma=4.0))

    def test_transpose_flip_handles_args(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(30, 60)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.transpose_flip(a.xdata, flip_v=True)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.flipud(d))

    def test_crop_handles_args(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            d = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region = Graphics.RectangleGraphic()
            region.center = 0.41, 0.51
            region.size = 0.52, 0.42
            display_item.add_graphic(region)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, regionA.bounds)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computation.create_input_item("regionA", Symbolic.make_item(region))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[9:42, 19:45])

    def test_evaluate_computation_gives_correct_value(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertTrue(numpy.array_equal(data_and_metadata.data, -data))

    def test_computation_fires_needs_update_event_when_data_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            with contextlib.closing(computation.computation_mutated_event.listen(needs_update)):
                with data_item.data_ref() as dr:
                    dr.data += 1.5
            self.assertTrue(needs_update_ref[0])

    def test_computation_fires_needs_update_event_when_display_data_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.uniform(0, 10, (2, 2, 2, 2)).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.display_xdata"))
            computation.create_input_item("a", Symbolic.make_item(display_item.display_data_channel))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            # verify assumptions
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(data[0, 0, ...], computed_data_item.data))
            # change display, check
            display_item.display_data_channels[0].collection_index = 1, 1
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(data[1, 1, ...], computed_data_item.data))

    def test_computation_fires_needs_update_event_when_metadata_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            with contextlib.closing(computation.computation_mutated_event.listen(needs_update)):
                metadata = data_item.metadata
                metadata["abc"] = 1
                data_item.metadata = metadata
            self.assertTrue(needs_update_ref[0])

    def test_computation_does_not_update_when_graphic_changes_on_source(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.add_graphic(Graphics.PointGraphic())
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            with contextlib.closing(computation.computation_mutated_event.listen(needs_update)):
                display_item.graphics[0].position = (0.3, 0.4)
            self.assertFalse(needs_update_ref[0])

    def test_computation_fires_needs_update_event_when_object_property(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region = Graphics.RectangleGraphic()
            region.center = 0.41, 0.51
            region.size = 0.52, 0.42
            display_item.add_graphic(region)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, regionA.bounds)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computation.create_input_item("regionA", Symbolic.make_item(region))
            document_model.append_computation(computation)
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            with contextlib.closing(computation.computation_mutated_event.listen(needs_update)):
                display_item.graphics[0].size = 0.53, 0.43
            self.assertTrue(needs_update_ref[0])

    def test_computation_fires_needs_update_event_when_variable_or_object_added(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region = Graphics.RectangleGraphic()
            region.center = 0.41, 0.51
            region.size = 0.52, 0.42
            display_item.add_graphic(region)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + n"))
            document_model.append_computation(computation)
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            with contextlib.closing(computation.computation_mutated_event.listen(needs_update)):
                computation.create_input_item("a", Symbolic.make_item(data_item))
                self.assertTrue(needs_update_ref[0])
                needs_update_ref[0] = False
                computation.create_variable("x", value_type="integral", value=5)
                self.assertTrue(needs_update_ref[0])

    def test_computation_handles_data_lookups(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller_with_application()
            document_model = document_controller.document_model
            d = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-api.library.get_data_item_by_uuid(uuid.UUID('{}')).xdata".format(str(data_item.uuid))))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            assert numpy.array_equal(data_and_metadata.data, -d)

    def test_computation_handles_region_lookups(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller_with_application()
            document_model = document_controller.document_model
            d = numpy.random.randn(100, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region = Graphics.RectangleGraphic()
            region.center = 0.5, 0.5
            region.size = 0.6, 0.4
            display_item.add_graphic(region)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, api.library.get_graphic_by_uuid(uuid.UUID('{}')).bounds)".format(str(region.uuid))))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            assert numpy.array_equal(data_and_metadata.data, d[20:80, 30:70])

    def test_computation_does_not_copy_metadata_during_computation(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            data_item.metadata = {"abc": 1}
            data_item.set_intensity_calibration(Calibration.Calibration(1.0, 2.0, "nm"))
            data_item.set_dimensional_calibrations([Calibration.Calibration(1.1, 2.1, "m"), Calibration.Calibration(1.2, 2.2, "m")])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata / numpy.average(a.data) * 5"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(data_and_metadata.metadata, dict())
            self.assertEqual(data_and_metadata.intensity_calibration, data_item.intensity_calibration)
            self.assertEqual(data_and_metadata.dimensional_calibrations, data_item.dimensional_calibrations)

    def test_remove_data_item_with_computation_succeeds(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            d = numpy.ones((8, 8), dtype=numpy.uint32)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            map = {"a": Symbolic.make_item(data_item)}
            new_data_item = document_controller.processing_computation("-a.xdata", map)
            document_model.recompute_all()
            document_model.remove_data_item(new_data_item)

    def test_evaluate_corrupt_computation_gives_sensible_response(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("(a.xdata++)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertIsNone(data_and_metadata)
            computation.close()

    def test_evaluate_computation_with_invalid_source_gives_sensible_response(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + e"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertIsNone(data_and_metadata)
            computation.close()

    def test_evaluate_computation_with_invalid_function_in_document_fails_cleanly(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            map = {"a": Symbolic.make_item(data_item)}
            document_controller.processing_computation("void(a,2)", map)
            document_model.recompute_all()

    def test_computation_changed_updates_evaluated_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertTrue(numpy.array_equal(data_and_metadata.data, -data))
            computation.expression = Symbolic.xdata_expression("-2 * a.xdata")
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertTrue(numpy.array_equal(data_and_metadata.data, -data*2))

    def test_changing_computation_updates_data_item(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            map = {"a": Symbolic.make_item(data_item)}
            computed_data_item = document_controller.processing_computation(Symbolic.xdata_expression("-a.xdata"), map)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data))
            document_model.get_data_item_computation(computed_data_item).expression = Symbolic.xdata_expression("-a.xdata * 2")
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data * 2))

    def test_unary_functions_return_correct_dimensions(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.data_expression("numpy.sin(a.data)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(len(data_and_metadata.dimensional_calibrations), 2)

    def test_computation_stores_original_text(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            data_expression = Symbolic.data_expression("numpy.sin(a.data)")
            computation = document_model.create_computation(data_expression)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            self.assertEqual(computation.expression, data_expression)

    def test_computation_stores_error_and_original_text(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            xdata_expression = Symbolic.xdata_expression("xyz(a.xdata)")
            computation = document_model.create_computation(xdata_expression)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertIsNone(data_and_metadata)
            self.assertTrue(computation.error_text is not None and len(computation.error_text) > 0)
            self.assertEqual(computation.expression, xdata_expression)

    def test_computation_reloads_missing_scalar_function(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.zeros((8, 8), dtype=numpy.uint32)
            src_data[:] = random.randint(0, 100)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("numpy.average(a.data)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_node_dict = computation.write_to_dict()
            data_node_dict['original_expression'] = "missing(a.xdata)"
            computation2 = document_model.create_computation()
            computation2.read_from_dict(data_node_dict)
            self.assertIsNone(DocumentModel.evaluate_data(computation2))
            computation2.close()

    def test_computation_can_extract_item_from_scalar_tuple(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.uniform(0, 10, (4, 2)).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + a.xdata.data_shape[1] + a.xdata.data_shape[0]"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertTrue(numpy.array_equal(data_and_metadata.data, data + 6))

    def test_columns_and_rows_and_radius_functions_return_correct_values(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.row(a.xdata.data_shape, -1, 1) + xd.column(a.xdata.data_shape, -1, 1) + xd.radius(a.xdata.data_shape)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            icol, irow = numpy.meshgrid(numpy.linspace(-1, 1, 8), numpy.linspace(-1, 1, 10))
            self.assertTrue(numpy.array_equal(data_and_metadata.data, icol + irow + numpy.sqrt(pow(icol, 2) + pow(irow, 2))))

    def test_copying_data_item_with_computation_copies_computation(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            copied_data_item = document_model.copy_data_item(computed_data_item)
            self.assertIsNotNone(document_model.get_data_item_computation(copied_data_item))
            self.assertEqual(document_model.get_data_item_computation(computed_data_item).error_text, document_model.get_data_item_computation(copied_data_item).error_text)
            self.assertEqual(document_model.get_data_item_computation(computed_data_item).expression, document_model.get_data_item_computation(copied_data_item).expression)

    def test_changing_computation_source_data_updates_computation(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(data_item.data, data))
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data_item.data))
            data_item.set_data(numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32))
            document_model.recompute_all()
            self.assertFalse(numpy.array_equal(data_item.data, data))
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data_item.data))

    def test_computation_is_live_after_copying_data_item_with_computation(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.int32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data_item.data))
            copied_data_item = document_model.copy_data_item(computed_data_item)
            data_item.set_data(((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.int32))
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data_item.data))
            self.assertTrue(numpy.array_equal(copied_data_item.data, -data_item.data))

    def test_computation_extracts_data_property_of_data_item(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, data_item.data))

    def test_resample_produces_correct_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.resample_image(a.xdata, (5, 4))"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, Image.scaled(data_item.data, (5, 4))))

    def test_resample_with_data_shape_produces_correct_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            data2 = numpy.zeros((5, 4), numpy.uint32)
            data_item2 = DataItem.DataItem(data2)
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.resample_image(a.xdata, b.xdata.data_shape)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computation.create_input_item("b", Symbolic.make_item(data_item2))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, Image.scaled(data_item.data, (5, 4))))

    def test_computation_extracts_display_data_property_of_data_item(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_data_channel = display_item.display_data_channel
            computation = document_model.create_computation(Symbolic.xdata_expression("a.display_xdata"))
            computation.create_input_item("a", Symbolic.make_item(display_data_channel))
            data_node_dict = computation.write_to_dict()
            data_node_dict["uuid"] = str(uuid.uuid4())
            data_node_dict["variables"][0]["uuid"] = str(uuid.uuid4())
            computation2 = document_model.create_computation()
            computation2.read_from_dict(data_node_dict)
            computation2.needs_update = True
            document_model.append_computation(computation)
            document_model.append_computation(computation2)
            data = DocumentModel.evaluate_data(computation).data
            data2 = DocumentModel.evaluate_data(computation2).data
            assert numpy.array_equal(data, src_data)
            assert numpy.array_equal(data, data2)

    def test_evaluation_with_variable_produces_correct_data(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            d = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d + 5)

    def test_evaluation_with_two_variables_produces_correct_data(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, x - y)"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_variable("y", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            assert numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data)

    def test_changing_variable_value_updates_computation(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            x = computation.create_variable("x", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data + 5))
            x.value = 8
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data + 8))

    def test_changing_region_property_updates_computation(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            src_data = numpy.random.randn(20, 20)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            line_region = Graphics.LineProfileGraphic()
            line_region.start = 0.25, 0.25
            line_region.end = 0.75, 0.75
            display_item.add_graphic(line_region)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.line_profile(src.display_xdata, line_region.vector, line_region.line_width)"))
            computation.create_input_item("src", Symbolic.make_item(display_item.display_data_channel))
            computation.create_input_item("line_region", Symbolic.make_item(line_region))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, Core.function_line_profile(data_item.xdata, line_region.vector, 1.0).data))
            line_region.start = 0.25, 0.20
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, Core.function_line_profile(data_item.xdata, line_region.vector, 1.0).data))

    def test_changing_variable_name_has_no_effect_on_computation(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            x = computation.create_variable("x", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data + 5))
            x.name = "xx"
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data + 5))

    def test_computation_with_variable_reloads(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            d = computation.write_to_dict()
            d["uuid"] = str(uuid.uuid4())
            d["variables"][0]["uuid"] = str(uuid.uuid4())
            d["variables"][1]["uuid"] = str(uuid.uuid4())
            computation2 = document_model.create_computation()
            computation2.read_from_dict(d)
            computation2.needs_update = True
            document_model.append_computation(computation)
            document_model.append_computation(computation2)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation2).data, src_data + 5))

    def test_computation_variable_writes_and_reads(self):
        with contextlib.closing(Symbolic.ComputationVariable("x", value_type="integral", value=5)) as variable:
            self.assertEqual(variable.name, "x")
            self.assertEqual(variable.value, 5)
            data_node_dict = variable.write_to_dict()
            with contextlib.closing(Symbolic.ComputationVariable()) as variable2:
                variable2.read_from_dict(data_node_dict)
                self.assertEqual(variable.name, variable2.name)
                self.assertEqual(variable.value, variable2.value)

    def test_computation_variable_change_type(self):
        with contextlib.closing(Symbolic.ComputationVariable("x", value_type="integral", value=5)) as variable:
            variable.variable_type = "data_item"
        with contextlib.closing(Symbolic.ComputationVariable("x", value_type="integral", value=5)) as variable:
            variable.variable_type = "graphic"

    def test_computation_reparsing_keeps_variables(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            x = computation.create_variable("x", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            computation.expression = Symbolic.xdata_expression("x + a.xdata")
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))

    def test_computation_using_object_parses_and_evaluates(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))

    def test_computation_using_object_updates_when_data_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            d = computation.write_to_dict()
            d["uuid"] = str(uuid.uuid4())
            d["variables"][0]["uuid"] = str(uuid.uuid4())
            d["variables"][1]["uuid"] = str(uuid.uuid4())
            read_computation = document_model.create_computation()
            read_computation.read_from_dict(d)
            read_computation.needs_update = True
            document_model.append_computation(read_computation)
            src_data2 = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item.set_data(src_data2)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(read_computation).data, src_data2 + 5))

    def test_computation_using_object_updates_efficiently_when_region_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region = Graphics.RectangleGraphic()
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.5, 0.5), Geometry.FloatSize(0.5, 0.5))
            display_item.add_graphic(region)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, r.bounds)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computation.create_input_item("r", Symbolic.make_item(region))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[3:9, 2:6]))
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.25, 0.25), Geometry.FloatSize(0.5, 0.5))
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.0, 0.0), Geometry.FloatSize(0.5, 0.5))
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.25, 0.25), Geometry.FloatSize(0.5, 0.5))
            evaluation_count = computation._evaluation_count_for_test
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[0:6, 0:4]))

    def test_computation_updates_efficiently_when_variable_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region = Graphics.RectangleGraphic()
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.5, 0.5), Geometry.FloatSize(0.5, 0.5))
            display_item.add_graphic(region)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, s)"))
            s = computation.create_variable("s", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            evaluation_count = computation._evaluation_count_for_test
            s.value = 4
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 0)
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)

    def test_computation_updates_efficiently_when_variable_added_or_removed(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region = Graphics.RectangleGraphic()
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.5, 0.5), Geometry.FloatSize(0.5, 0.5))
            display_item.add_graphic(region)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, s)"))
            s = computation.create_variable("s", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            evaluation_count = computation._evaluation_count_for_test
            t = computation.create_variable("t", value_type="integral", value=5)
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 0)
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)
            computation.remove_variable(t)
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 2)

    def test_computation_updates_efficiently_when_expression_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region = Graphics.RectangleGraphic()
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.5, 0.5), Geometry.FloatSize(0.5, 0.5))
            display_item.add_graphic(region)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, s)"))
            s = computation.create_variable("s", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            evaluation_count = computation._evaluation_count_for_test
            computation.expression = Symbolic.xdata_expression("xd.gaussian_blur(a, 2)")
            # computation should not be re-evaluated until requested
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 0)
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)

    def test_computation_with_object_writes_and_reads(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + 4"))
            a_specifier = computation.create_input_item("a", Symbolic.make_item(data_item)).specifier
            d = computation.write_to_dict()
            document_model.append_computation(computation)
            read_computation = document_model.create_computation()
            read_computation.read_from_dict(d)
            self.assertEqual(read_computation.variables[0].name, "a")
            self.assertEqual(read_computation.variables[0].specifier, a_specifier)
            read_computation.close()

    def test_computation_with_object_reloads(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            d = computation.write_to_dict()
            d["uuid"] = str(uuid.uuid4())
            d["variables"][0]["uuid"] = str(uuid.uuid4())
            d["variables"][1]["uuid"] = str(uuid.uuid4())
            computation2 = document_model.create_computation()
            computation2.read_from_dict(d)
            computation2.needs_update = True
            document_model.append_computation(computation)
            document_model.append_computation(computation2)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation2).data, src_data + 5))

    def test_computation_with_object_evaluates_correctly_after_changing_the_variable_name(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            x = computation.create_variable("x", value_type="integral", value=5)
            document_model.append_computation(computation)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            x.name = "xx"
            computation.expression = Symbolic.xdata_expression("a.xdata + xx")
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))

    def test_computation_with_object_evaluates_correctly_after_changing_the_specifier(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data1 = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            src_data2 = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item1 = DataItem.DataItem(src_data1)
            data_item2 = DataItem.DataItem(src_data2)
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            expression = "a.xdata + 1"
            computation = document_model.create_computation(Symbolic.xdata_expression(expression))
            a = computation.create_input_item("a", Symbolic.make_item(data_item1))
            document_model.append_computation(computation)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data1 + 1))
            computation.set_input_item("a", Symbolic.make_item(data_item2))
            computation.expression = Symbolic.xdata_expression(expression)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data2 + 1))

    def test_computation_fires_needs_update_event_when_specifier_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data1 = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            src_data2 = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item1 = DataItem.DataItem(src_data1)
            data_item2 = DataItem.DataItem(src_data2)
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            expression = "a.xdata + 1"
            computation = document_model.create_computation(Symbolic.xdata_expression(expression))
            a = computation.create_input_item("a", Symbolic.make_item(data_item1))
            document_model.append_computation(computation)
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            with contextlib.closing(computation.computation_mutated_event.listen(needs_update)):
                computation.set_input_item("a", Symbolic.make_item(data_item2))
            self.assertTrue(needs_update_ref[0])

    def test_computation_in_document_recomputes_when_specifier_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data1 = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            src_data2 = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item1 = DataItem.DataItem(src_data1)
            data_item2 = DataItem.DataItem(src_data2)
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            expression = "a.xdata + 1"
            computation = document_model.create_computation(Symbolic.xdata_expression(expression))
            a = computation.create_input_item("a", Symbolic.make_item(data_item1))
            computed_data_item = DataItem.DataItem(src_data1.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data1 + 1))
            computation.set_input_item("a", Symbolic.make_item(data_item2))
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data2 + 1))

    def test_computation_in_document_is_still_live_when_region_specifier_uuid_str_changes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(10, 10)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region1 = Graphics.RectangleGraphic()
            region1.bounds = Geometry.FloatRect.from_tlhw(0.0, 0.0, 0.5, 0.5)
            display_item.add_graphic(region1)
            region2 = Graphics.RectangleGraphic()
            region2.bounds = Geometry.FloatRect.from_tlhw(0.5, 0.5, 0.5, 0.5)
            display_item.add_graphic(region2)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, r.bounds)"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            r = computation.create_input_item("r", Symbolic.make_item(region1))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            # verify assumptions
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[0:5, 0:5]))
            # now switch the region uuid
            # TODO: should this happen automatically without the unbind/bind?
            r.unbind()
            r.specifier.reference = region2
            r.bind()
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[5:10, 5:10]))
            # and make sure recompute happens when new region uuid changes
            region2.bounds = Geometry.FloatRect.from_tlhw(0.0, 0.0, 0.5, 0.5)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[0:5, 0:5]))

    def test_computation_with_raw_reference_copies(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = numpy.random.uniform(0, 10, (10, 8)).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            region = Graphics.RectangleGraphic()
            display_item.add_graphic(region)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            a = computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            computation.remove_variable(a)
            copy.deepcopy(computation).close()

    def test_evaluation_error_recovers_gracefully(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.line_profile(a.xdata, xd.vector(xd.norm_point(0.25, 0.25), xd.norm_point(0.5, 0.5)), x)"))
            x = computation.create_variable("x", value_type="integral", value=0)
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertIsNotNone(computation.error_text)
            self.assertEqual(len(computed_data_item.data.shape), 2)  # original data
            x.value = 1
            document_model.recompute_all()
            self.assertIsNone(computation.error_text)
            self.assertIsNotNone(computed_data_item.data)
            self.assertEqual(len(computed_data_item.data.shape), 1)  # computed data

    def test_various_expressions_produces_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            script_and_data = [
                ("xd.histogram(a.xdata, 10)", None),
                ("xd.line_profile(a.xdata, xd.vector(xd.norm_point(0.1, 0.1), xd.norma_point(0.8, 0.7)), 10)", None),
                ("xd.transpose_flip(a.xdata, False, True, False)", None),
                ("xd.crop(a.xdata, xd.rectangle_from_origin_size(xd.norm_point(0, 0), xd.norm_size(0.5, 0.625)))", src_data[0:5, 0:5]),
                ("xd.sum(a.xdata, 0)", numpy.sum(src_data, 0)),
                ("xd.resample_image(a.xdata, (32, 32))", Image.scaled(src_data, (32, 32))),
                ("xd.resample_image(a.xdata, a.xdata.data_shape)", src_data),
                ("xd.resample_image(a.xdata, xd.crop(a.xdata, xd.rectangle_from_origin_size(xd.norm_point(0, 0), xd.norm_size(0.5, 0.625))).data_shape)", Image.scaled(src_data, (5, 5))),
                ("a.xdata + x", src_data + 5),
                ("xd.gaussian_blur(a.xdata, x + x)", None),
                ("xd.gaussian_blur(a.xdata, x - 2)", None),
                ("xd.gaussian_blur(a.xdata, 2 * x)", None),
                ("xd.gaussian_blur(a.xdata, +x)", None),
            ]
            for script, data in script_and_data:
                computation = document_model.create_computation(Symbolic.xdata_expression(script))
                computation.create_input_item("a", Symbolic.make_item(data_item))
                computation.create_variable("x", value_type="integral", value=5)
                computed_data_item = DataItem.DataItem(src_data.copy())
                document_model.append_data_item(computed_data_item)
                document_model.set_data_item_computation(computed_data_item, computation)
                document_model.recompute_all()
                self.assertIsNotNone(computed_data_item.data)
                if data is not None:
                    self.assertTrue(numpy.array_equal(computed_data_item.data, data))

    def test_conversion_to_int(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.float64)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation()
            computation.create_input_item("src", Symbolic.make_item(data_item))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, int)")
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.int_)
            self.assertTrue(numpy.array_equal(data, src_data.astype(int)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.int16)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.int16)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.int16)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.int32)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.int32)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.int32)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.int64)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.int64)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.int64)))

    def test_conversion_to_uint(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.float64)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation()
            computation.create_input_item("src", Symbolic.make_item(data_item))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.uint8)")
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.uint8)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.uint8)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.uint16)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.uint16)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.uint16)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.uint32)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.uint32)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.uint32)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.uint64)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.uint64)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.uint64)))

    def test_conversion_to_float(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.int32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation()
            computation.create_input_item("src", Symbolic.make_item(data_item))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.float32)")
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.float32)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.float32)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.float64)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.float64)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.float64)))

    def test_conversion_to_complex(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.int32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation()
            computation.create_input_item("src", Symbolic.make_item(data_item))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.complex64)")
            document_model.append_computation(computation)
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.complex64)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.complex64)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.complex128)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.complex128)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.complex128)))

    def test_data_descriptor_is_maintained_during_evaluate(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            src_data1 = ((numpy.abs(numpy.random.randn(10,)) + 1) * 10).astype(numpy.int32)
            data_item1 = DataItem.DataItem(src_data1)
            document_model.append_data_item(data_item1)
            src_data2 = ((numpy.abs(numpy.random.randn(10,)) + 1) * 10).astype(numpy.int32)
            data_item2 = DataItem.DataItem(src_data2)
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation()
            computation.create_input_item("src1", Symbolic.make_item(data_item1))
            computation.create_input_item("src2", Symbolic.make_item(data_item2))
            computation.expression = Symbolic.xdata_expression("xd.vstack((src1.xdata, src2.xdata))")
            document_model.append_computation(computation)
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(data_and_metadata.collection_dimension_count, 1)
            self.assertEqual(data_and_metadata.datum_dimension_count, 1)

    def test_computation_evaluates_on_thread(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            # in order to make this test go fast, attach the call_soon event here
            # and handle it, setting the continue event at the same time.
            # this must go before the document controller is created since call_soon
            # is a 'fire_any' style event, meaning that only the first handler takes
            # it. we return False to let the document controller eventually handle it.
            continue_event = threading.Event()
            def do_call_soon():
                continue_event.set()
                return False
            listener = document_model.call_soon_event.listen(do_call_soon)
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.start_dispatcher()
            continue_event.wait(10.0)
            listener.close()
            listener = None
            document_controller.periodic()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -src_data))

    def test_computation_on_deleted_data_item(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.remove_data_item(data_item)
            document_model.recompute_all()

    def test_computation_updates_timezone(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            try:
                src_data = numpy.random.randn(2, 2)
                data_item = DataItem.DataItem(src_data)
                document_model.append_data_item(data_item)
                computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
                computation.create_input_item("a", Symbolic.make_item(data_item))
                Utility.local_timezone_override = [None]
                Utility.local_utcoffset_override = [0]
                computed_data_item = DataItem.DataItem(src_data.copy())
                document_model.append_data_item(computed_data_item)
                document_model.set_data_item_computation(computed_data_item, computation)
                self.assertFalse(computed_data_item.timezone)
                self.assertEqual(computed_data_item.timezone_offset, "+0000")
                Utility.local_timezone_override = ["Europe/Athens"]
                Utility.local_utcoffset_override = [180]
                document_model.recompute_all()
                self.assertEqual(computed_data_item.timezone, "Europe/Athens")
                self.assertEqual(computed_data_item.timezone_offset, "+0300")
            finally:
                Utility.local_timezone_override = None
                Utility.local_utcoffset_override = None

    def test_deleting_one_variable_on_bound_computation_rebinds_properly(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("x"))
            x_var = computation.create_variable("x")
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            computation.remove_variable(x_var)

    class ComputeExecError:
        def __init__(self, computation, **kwargs):
            self.computation = computation

        def execute(self, src_xdata):
            raise RuntimeError()

        def commit(self):
            pass

    def test_computation_error_is_handled_in_main_thread(self):
        Symbolic.register_computation_type("compute_error", self.ComputeExecError)
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            # in order to make this test go fast, attach the call_soon event here
            # and handle it, setting the continue event at the same time.
            # this must go before the document controller is created since call_soon
            # is a 'fire_any' style event, meaning that only the first handler takes
            # it. we return False to let the document controller eventually handle it.
            continue_event = threading.Event()
            def do_call_soon():
                continue_event.set()
                return False
            listener = document_model.call_soon_event.listen(do_call_soon)
            data_item = DataItem.DataItem(numpy.zeros((2, 2)))
            document_model.append_data_item(data_item)
            computation1 = document_model.create_computation()
            computation1.create_input_item("src_xdata", Symbolic.make_item(data_item, type="xdata"))
            computation1.processing_id = "compute_error"
            document_model.append_computation(computation1)
            document_model.start_dispatcher()
            continue_event.wait(10.0)
            listener.close()
            document_controller.periodic()

    def test_computation_error_clear_initial_computation_flag(self):
        Symbolic.register_computation_type("compute_error", self.ComputeExecError)
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((2, 2)))
            document_model.append_data_item(data_item)
            computation1 = document_model.create_computation()
            computation1.create_input_item("src_xdata", Symbolic.make_item(data_item, type="xdata"))
            computation1.processing_id = "compute_error"
            document_model.append_computation(computation1)
            document_model.recompute_all()
            self.assertTrue(computation1.is_initial_computation_complete.wait(0.01))

    class ComputeExecDelayedError:
        def __init__(self, started_event: threading.Event, continue_event: threading.Event, computation, **kwargs):
            self.computation = computation
            self.started_event = started_event
            self.continue_event = continue_event

        def execute(self, src_xdata):
            self.started_event.set()
            self.continue_event.wait(5)
            raise RuntimeError()

        def commit(self):
            pass

    def test_computation_removed_during_execute_is_handled(self):
        started_event = threading.Event()
        continue_event = threading.Event()
        import functools
        Symbolic.register_computation_type("compute_delayed_error", functools.partial(self.ComputeExecDelayedError, started_event, continue_event))
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.zeros((2, 2)))
            document_model.append_data_item(data_item)
            computation1 = document_model.create_computation()
            computation1.create_input_item("src_xdata", Symbolic.make_item(data_item, type="xdata"))
            computation1.processing_id = "compute_delayed_error"
            document_model.append_computation(computation1)
            document_model.start_dispatcher()
            started_event.wait(5)
            # computation will be in execute now
            document_model.remove_computation(computation1)
            # now let the computation finish
            continue_event.set()
            # give it a chance to fully complete with a pending error.
            time.sleep(0.05)
            document_model.recompute_all()
            document_controller.periodic()

    def test_removing_computation_unbinds(self):
        Symbolic.register_computation_type("compute_error", self.ComputeExecError)
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.random.randn(2, 2))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            document_model.get_invert_new(display_item, display_item.data_item)
            computation = document_model.computations[0]
            self.assertIsNotNone(computation.get_input("src"))
            document_model.remove_data_item(document_model.data_items[1])
            self.assertTrue(computation._closed)

    def test_removing_data_item_source_data_item_from_library_is_possible(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((2, 2)))
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata"))
            computation.create_input_item("a", Symbolic.make_item(data_item))
            document_model.append_computation(computation)
            self.assertTrue(computation.is_resolved)
            document_model.remove_data_item(data_item)
            self.assertTrue(computation._closed)

    def test_adjusting_interval_on_line_profile_does_not_trigger_recompute(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_display_item(display_item)
            line_profile_display_item = document_controller.processing_line_profile()
            document_model.recompute_all()
            self.assertEqual(1, document_model.computations[-1]._evaluation_count_for_test)
            interval_region = Graphics.IntervalGraphic()
            line_profile_display_item.add_graphic(interval_region)
            document_model.recompute_all()
            self.assertEqual(1, document_model.computations[-1]._evaluation_count_for_test)
            interval_region.interval = 0.2, 0.3
            document_model.recompute_all()
            self.assertEqual(1, document_model.computations[-1]._evaluation_count_for_test)

    def test_data_source_watches_correct_graphics(self):
        # this failed at one point due to improper use of local variable
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem()
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            crop_region = Graphics.RectangleGraphic()
            display_item.add_graphic(crop_region)
            graphic = Graphics.PointGraphic()
            display_item.add_graphic(graphic)
            document_model.get_fft_new(display_item, display_item.data_item, crop_region)
            display_item.remove_graphic(graphic).close()
            document_model.recompute_all()
            graphic = Graphics.PointGraphic()
            display_item.add_graphic(graphic)
            # at this stage the data source monitor was corrupt
            graphic.position = 0.2, 0.2  # this triggered an exception

    def disabled_test_reshape_rgb(self):
        assert False

    def disabled_test_computation_with_data_error_gets_reported(self):
        assert False  # when the data node returns None

    def disabled_test_computation_variable_gets_closed(self):
        assert False

    def disabled_test_computation_with_cycles_fails_gracefully(self):
        assert False

    def disabled_test_computations_handle_constant_values_as_errors(self):
        # computation.parse_expression(document_model, "7", dict())
        assert False

    def disabled_test_computations_update_data_item_dependencies_list(self):
        assert False

    def disabled_test_function_to_modify_intensity_calibration(self):
        assert False

    def disabled_test_function_to_modify_dimensional_calibrations(self):
        assert False

    def disabled_test_function_to_modify_metadata(self):
        assert False

    def disabled_test_data_slice_calibration_with_step(self):
        # d[::2, :, :]
        pass

    def disabled_test_invalid_computation_produces_error_message(self):
        # example d[3:3, ...]
        # should return None be allowed? what about raise exception in all cases?
        pass

if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



from __future__ import annotations

# standard libraries
import asyncio
import concurrent.futures
import gettext
import json
import threading
import time
import typing
import weakref

# third party libraries
# None

# local libraries
from nion.swift import Panel
from nion.ui import CanvasItem
from nion.ui import DrawingContext
from nion.ui import UserInterface
from nion.ui import TreeCanvasItem
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import Process

if typing.TYPE_CHECKING:
    from nion.data import DataAndMetadata
    from nion.swift import DocumentController
    from nion.swift.model import DataItem
    from nion.swift.model import DisplayItem

_ = gettext.gettext


class MetadataModel:
    """Represents metadata. Tracks a display specifier for changes to it and its metadata content.

    Provides read/write access to metadata via the property.

    Provides a metadata_changed event, always fired on UI thread.
    """

    def __init__(self, document_controller: DocumentController.DocumentController) -> None:
        self.__weak_document_controller = weakref.ref(document_controller)
        self.__data_item: typing.Optional[DataItem.DataItem] = None

        # thread safe.
        def display_item_changed(display_item: typing.Optional[DisplayItem.DisplayItem]) -> None:
            data_item = display_item.data_item if display_item else None

            def update_data_item() -> None:
                self.__set_data_item(data_item)

            document_controller = self.document_controller
            if document_controller:
                document_controller.add_task("update_data_item" + str(id(self)), update_data_item)

        self.__display_item_changed_event_listener = document_controller.focused_display_item_changed_event.listen(display_item_changed)
        self.__set_data_item(None)
        self.__property_changed_event_listener: typing.Optional[Event.EventListener] = None
        self.metadata_changed_event = Event.Event()

    def close(self) -> None:
        document_controller = self.document_controller
        if document_controller:
            document_controller.clear_task("update_data_item" + str(id(self)))
        self.__display_item_changed_event_listener.close()
        self.__display_item_changed_event_listener = typing.cast(typing.Any, None)
        self.__set_data_item(None)

    @property
    def document_controller(self) -> typing.Optional[DocumentController.DocumentController]:
        return self.__weak_document_controller()

    def __metadata_changed(self, data_item: typing.Optional[DataItem.DataItem]) -> None:
        self.metadata_changed_event.fire(data_item)

    # not thread safe
    def __set_data_item(self, data_item: typing.Optional[DataItem.DataItem]) -> None:
        if self.__data_item != data_item:
            if self.__property_changed_event_listener:
                self.__property_changed_event_listener.close()
                self.__property_changed_event_listener = None
            self.__data_item = data_item
            # update the expression text
            if data_item:
                def property_changed(property_name: str) -> None:
                    if property_name == "metadata":
                        self.__metadata_changed(data_item)
                self.__property_changed_event_listener = data_item.property_changed_event.listen(property_changed)
            self.__metadata_changed(data_item)


class MetadataSource(typing.Protocol):
    @property
    def metadata(self) -> DataAndMetadata.MetadataType:
        raise NotImplementedError()


class MetadataEditorTreeDelegate(TreeCanvasItem.TreeCanvasItemDelegate):
    def __init__(self) -> None:
        self.metadata_source: typing.Optional[MetadataSource] = None
        self.__expanded_value_paths: typing.Set[str] = set()

    def __is_expanded(self, value_path: TreeCanvasItem._ValuePath) -> bool:
        return json.dumps(value_path) in self.__expanded_value_paths

    def toggle_is_expanded(self, value_path_key: str) -> None:
        if value_path_key in self.__expanded_value_paths:
            self.__expanded_value_paths.remove(value_path_key)
        else:
            self.__expanded_value_paths.add(value_path_key)

    def build_items(self, get_font_metrics_fn: typing.Callable[[str, str], UserInterface.FontMetrics],
                    item_width: typing.Optional[int]) -> typing.Sequence[TreeCanvasItem.TreeItem]:
        items: typing.List[TreeCanvasItem.TreeItem] = list()
        text_font = "normal 12px monospace"

        def visit_value(value_path: TreeCanvasItem._ValuePath, value: typing.Any) -> None:
            if isinstance(value, dict):
                is_expanded = self.__is_expanded(value_path)
                format_str = "{} {{{}}}"
                text_item = CanvasItem.StaticTextCanvasItem(format_str.format(value_path[-1], len(value)))
                text_item.font = text_font
                text_item.size_to_content(get_font_metrics_fn)
                items.append(TreeCanvasItem.TreeItem(text_item, "parent", is_expanded, value_path))
                if is_expanded:
                    visit_dict(value, value_path)
            elif isinstance(value, list) or isinstance(value, tuple):
                is_expanded = self.__is_expanded(value_path)
                format_str = "{} ({})"
                text_item = CanvasItem.StaticTextCanvasItem(format_str.format(value_path[-1], len(value)))
                text_item.font = text_font
                text_item.size_to_content(get_font_metrics_fn)
                items.append(TreeCanvasItem.TreeItem(text_item, "parent", is_expanded, value_path))
                if is_expanded:
                    visit_list(value, value_path)
            else:
                text_item = CanvasItem.StaticTextCanvasItem("{}: {}".format(value_path[-1], value))
                text_item.font = text_font
                text_item.size_to_content(get_font_metrics_fn)
                items.append(TreeCanvasItem.TreeItem(text_item, "child", False, value_path))

        def visit_list(l: typing.Sequence[typing.Any], path: TreeCanvasItem._ValuePath) -> None:
            for index, value in enumerate(l):
                value_path = tuple(path) + (index,)
                visit_value(value_path, value)

        def visit_dict(d: typing.Mapping[str, typing.Any], path: TreeCanvasItem._ValuePath) -> None:
            for key in sorted(d.keys()):
                value = d[key]
                value_path = tuple(path) + (key,)
                visit_value(value_path, value)

        if self.metadata_source:
            visit_dict(self.metadata_source.metadata, tuple())

        return items


class ThreadedCanvasItem(CanvasItem.CanvasItemComposition):
    _executor = concurrent.futures.ThreadPoolExecutor()

    def __init__(self, get_font_metrics_fn: typing.Callable[[str, str], UserInterface.FontMetrics], delegate: TreeCanvasItem.TreeCanvasItemDelegate, content_size_changed_fn: typing.Optional[typing.Callable[[Geometry.IntSize], None]] = None) -> None:
        super().__init__()
        self.__get_font_metrics_fn = get_font_metrics_fn
        self.__delegate = delegate
        self.__metadata_editor_canvas_item: typing.Optional[TreeCanvasItem.TreeCanvasItem] = None
        self.__thread_lock = threading.Lock()
        self.__thread_future: typing.Optional[concurrent.futures.Future[None]] = None
        self.__closing = False
        self.__pending = False
        self.__visible = False
        self.on_content_size_changed = content_size_changed_fn

        self.__draw(DrawingContext.DrawingContext())

        self.wants_mouse_events = True

    def close(self) -> None:
        # wait for display values threads to finish. first notify the thread that we are closing, then wait for it
        # to complete by getting the future and waiting for it to complete. then clear the streams to release any
        # resources (display values).
        self.__closing = True
        with self.__thread_lock:
            thread_future = self.__thread_future
        if thread_future:
            thread_future.result()
        super().close()

    def __draw_thread(self) -> None:
        while not self.__closing:
            with self.__thread_lock:
                pending = self.__pending
                self.__pending = False
            if pending:
                time.sleep(0.05)
                drawing_context = DrawingContext.DrawingContext()
                if not self.__closing:  # cleaner closing behavior, but not perfect.
                    with Process.audit("draw_thread"):
                        self.__draw(drawing_context)
            with self.__thread_lock:
                if not self.__pending or not self.__visible:
                    self.__thread_future = None
                    break

    def __draw(self, drawing_context: DrawingContext.DrawingContext) -> None:
        canvas_bounds = self.canvas_bounds
        if canvas_bounds:
            # start = time.perf_counter_ns()

            metadata_editor_canvas_item = TreeCanvasItem.TreeCanvasItem(self.__get_font_metrics_fn, self.__delegate)
            metadata_editor_canvas_item.on_reconstruct = self._trigger
            metadata_editor_canvas_item.reconstruct()
            # calculate the new preferred size of the metadata editor.
            metadata_editor_canvas_item.size_to_content()
            metadata_editor_canvas_item_size = Geometry.FloatSize(w=metadata_editor_canvas_item.sizing.preferred_width, h=metadata_editor_canvas_item.sizing.preferred_height).to_int_size()

            # modifying the composition is safe until it is added to this composition.
            canvas_item_composition = CanvasItem.CanvasItemComposition()
            canvas_item_composition.wants_mouse_events = True
            canvas_item_composition.layout = CanvasItem.CanvasItemColumnLayout()
            canvas_item_composition.add_canvas_item(metadata_editor_canvas_item)
            canvas_item_composition.add_stretch()
            self.__metadata_editor_canvas_item = metadata_editor_canvas_item

            # the canvas_bounds at this point does not represent the size of the reconstructed metadata editor.
            # so use the explicit size of the metadata editor to lay out the new composition.
            canvas_item_composition.update_layout(self.canvas_origin, metadata_editor_canvas_item_size)
            canvas_item_composition.repaint_immediate(drawing_context, metadata_editor_canvas_item_size)

            canvas_item_composition._set_owner_thread(threading.main_thread())

            self.replace_canvas_items([canvas_item_composition])

            if callable(self.on_content_size_changed):
                if metadata_editor_canvas_item.canvas_size:
                    self.on_content_size_changed(metadata_editor_canvas_item.canvas_size)

            # end = time.perf_counter_ns()
            # print(f"reconstruct {(end - start) / 1000}us")

    def _trigger(self) -> None:
        with self.__thread_lock:
            self.__pending = True  # pending even if not visible.
            if self.root_container and (canvas_widget := getattr(self.root_container, "canvas_widget", None)):
                self.__visible = typing.cast(UserInterface.DockWidget, typing.cast(UserInterface.CanvasWidget, canvas_widget).root_container).visible
                if self.__visible:
                    # only launch thread if visible and not already running.
                    if not self.__thread_future:
                        self.__thread_future = ThreadedCanvasItem._executor.submit(self.__draw_thread)

    # def _repaint_template(self, drawing_context: DrawingContext.DrawingContext, immediate: bool) -> None:
    #     start = time.perf_counter_ns()
    #     super()._repaint_template(drawing_context, immediate)
    #     end = time.perf_counter_ns()
    #     print(f"repaint {(end - start) / 1000}us")


class ThreadHelper:
    def __init__(self, event_loop: asyncio.AbstractEventLoop) -> None:
        self.__event_loop = event_loop
        self.__pending_calls: typing.Dict[str, asyncio.Handle] = dict()

    def close(self) -> None:
        for handle in self.__pending_calls.values():
            handle.cancel()
        self.__pending_calls = dict()

    def call_on_main_thread(self, key: str, func: typing.Callable[[], None]) -> None:
        if threading.current_thread() != threading.main_thread():
            handle = self.__pending_calls.pop(key, None)
            if handle:
                handle.cancel()
            def audited_func() -> None:
                with Process.audit(f"threadhelper.{key}"):
                    func()
            self.__pending_calls[key] = self.__event_loop.call_soon_threadsafe(audited_func)
        else:
            func()


class MetadataPanel(Panel.Panel):
    """Provide a panel to edit metadata."""

    def __init__(self, document_controller: DocumentController.DocumentController, panel_id: str, properties: typing.Mapping[str, typing.Any]) -> None:
        super().__init__(document_controller, panel_id, _("Metadata"))

        ui = self.ui

        self.__metadata_model = MetadataModel(document_controller)
        self.__thread_helper = ThreadHelper(document_controller.event_loop)

        delegate = MetadataEditorTreeDelegate()

        def content_size_changed(content_size: Geometry.IntSize) -> None:
            def _content_height_changed() -> None:
                # the code below will do the upstream part of the line below. we can't use the line below
                # because it will lay out the children again.
                # metadata_editor_canvas_item.update_layout(Geometry.IntPoint(), content_size)

                # leave the canvas origin alone so the scroll bars are at the same place
                # metadata_editor_canvas_item._set_canvas_origin(Geometry.IntPoint())

                # set the content size and send it on up so the scroll area sees it.
                metadata_editor_canvas_item._set_canvas_size(content_size)
            self.__thread_helper.call_on_main_thread("_content_height_changed", _content_height_changed)

        metadata_editor_canvas_item = ThreadedCanvasItem(ui.get_font_metrics, delegate, content_size_changed)

        self.__metadata_editor_canvas_item = metadata_editor_canvas_item

        scroll_area_canvas_item = CanvasItem.ScrollAreaCanvasItem(metadata_editor_canvas_item)
        scroll_group_canvas_item = CanvasItem.CanvasItemComposition()
        vertical_scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(scroll_area_canvas_item)
        horizontal_scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(scroll_area_canvas_item, CanvasItem.Orientation.Horizontal)
        scroll_group_canvas_item.layout = CanvasItem.CanvasItemGridLayout(Geometry.IntSize(width=2, height=2))
        scroll_group_canvas_item.add_canvas_item(scroll_area_canvas_item, Geometry.IntPoint(x=0, y=0))
        scroll_group_canvas_item.add_canvas_item(vertical_scroll_bar_canvas_item, Geometry.IntPoint(x=1, y=0))
        scroll_group_canvas_item.add_canvas_item(horizontal_scroll_bar_canvas_item, Geometry.IntPoint(x=0, y=1))

        metadata_editor_widget = ui.create_canvas_widget()
        metadata_editor_widget.canvas_item.layout = CanvasItem.CanvasItemColumnLayout()
        metadata_editor_widget.canvas_item.add_canvas_item(scroll_group_canvas_item)

        column = self.ui.create_column_widget(properties={"size-policy-horizontal": "expanding", "size-policy-vertical": "expanding"})
        column.add(metadata_editor_widget)

        def metadata_source_changed(metadata_source: MetadataSource) -> None:
            delegate.metadata_source = metadata_source

            def reconstruct_metadata() -> None:
                if self.__metadata_editor_canvas_item:  # use this instead of local variable to handle close properly
                    self.__metadata_editor_canvas_item._trigger()

            self.document_controller.queue_task(reconstruct_metadata)

        self.__metadata_changed_event_listener = self.__metadata_model.metadata_changed_event.listen(metadata_source_changed)

        self.widget = column

    def close(self) -> None:
        self.__thread_helper.close()
        self.__thread_helper = typing.cast(typing.Any, None)
        self.__metadata_editor_canvas_item = typing.cast(typing.Any, None)
        self.__metadata_changed_event_listener.close()
        self.__metadata_changed_event_listener = typing.cast(typing.Any, None)
        self.__metadata_model.close()
        self.__metadata_model = typing.cast(typing.Any, None)
        super().close()

    @property
    def _metadata_editor_canvas_item_for_testing(self) -> TreeCanvasItem.TreeCanvasItem:
        return typing.cast(TreeCanvasItem.TreeCanvasItem, self.__metadata_editor_canvas_item.canvas_items[0].canvas_items[0])

    @property
    def _metadata_model_for_testing(self) -> MetadataModel:
        return self.__metadata_model



"""
Provide a dialog to browse all project items.

Use the schema of the project to determine the types/structure of the project and allow the user to browser all
objects and follow links between objects.

Uses the general entity browser and registers factories for specific inspectors (with names ending in
'InspectorHandler') for each of the nionswift core objects with the entity browser.
"""

from __future__ import annotations

# standard libraries
import gettext
import operator
import typing

# third party libraries
# None

# local libraries
from nion.swift import ComputationPanel
from nion.swift import EntityBrowser
from nion.swift import Inspector
from nion.swift.model import DataItem
from nion.swift.model import DataStructure
from nion.swift.model import DisplayItem
from nion.swift.model import DocumentModel
from nion.swift.model import Model as DataModel
from nion.swift.model import Schema
from nion.swift.model import Symbolic
from nion.ui import Declarative
from nion.utils import Converter
from nion.utils import Registry

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController

_ = gettext.gettext


class DataItemInspectorHandler(Declarative.Handler):
    def __init__(self, context: EntityBrowser.Context, data_item: DataItem.DataItem):
        super().__init__()
        self.context = context
        self.item = data_item
        self.uuid_converter = Converter.UuidToStringConverter()
        self.date_converter = Converter.DatetimeToStringConverter(is_local=True, format="%Y-%m-%d %H:%M:%S %Z")
        u = Declarative.DeclarativeUI()
        label = u.create_label(text=self.item.title)
        uuid_row = u.create_row(u.create_label(text="UUID:", width=60), u.create_label(text="@binding(item.uuid, converter=uuid_converter)"), u.create_stretch(), spacing=12)
        modified_row = u.create_row(u.create_label(text="Modified:", width=60), u.create_label(text="@binding(item.modified, converter=date_converter)"), u.create_label(text="(local)"), u.create_stretch(), spacing=12)
        source_line = [u.create_component_instance("source_component")] if context.values.get("do_references", False) else []
        data_source_chooser = {
            "type": "data_source_chooser",
            "display_item": "@binding(display_item)",
            "min_width": 80,
            "min_height": 80,
        }
        self.ui_view = u.create_column(label,
                                       uuid_row,
                                       modified_row,
                                       *source_line,
                                       data_source_chooser,
                                       u.create_stretch(),
                                       spacing=12)

    @property
    def display_item(self) -> typing.Optional[DisplayItem.DisplayItem]:
        document_model = typing.cast(DocumentModel.DocumentModel, self.context.values["document_model"])
        return document_model.get_best_display_item_for_data_item(self.item)

    def create_handler(self, component_id: str, container: typing.Optional[Symbolic.ComputationVariable] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "source_component":
            return EntityBrowser.ReferenceHandler(self.context, _("Source"), self.item.source)
        return None


class DataItemInspectorHandlerFactory(EntityBrowser.ItemInspectorHandlerFactory):
    def make_sections(self, context: EntityBrowser.Context, item: typing.Any) -> typing.Sequence[typing.Tuple[str, Declarative.HandlerLike]]:
        if isinstance(item, DataItem.DataItem):
            return [(_("Data Item"), DataItemInspectorHandler(context, item))]
        return list()


class DisplayItemInspectorHandler(Declarative.Handler):
    def __init__(self, context: EntityBrowser.Context, display_item: DisplayItem.DisplayItem) -> None:
        super().__init__()

        self.item = display_item
        self.uuid_converter = Converter.UuidToStringConverter()
        self.date_converter = Converter.DatetimeToStringConverter(is_local=True, format="%Y-%m-%d %H:%M:%S %Z")

        u = Declarative.DeclarativeUI()

        label = u.create_label(text=display_item.title)
        uuid_row = u.create_row(u.create_label(text="UUID:", width=60),
                                u.create_label(text="@binding(item.uuid, converter=uuid_converter)"),
                                u.create_stretch(), spacing=12)
        modified_row = u.create_row(u.create_label(text="Modified:", width=60),
                                    u.create_label(text="@binding(item.modified, converter=date_converter)"),
                                    u.create_label(text="(local)"), u.create_stretch(), spacing=12)

        data_source_chooser = {
            "type": "data_source_chooser",
            "display_item": "@binding(item)",
            "min_width": 80,
            "min_height": 80,
        }

        self.ui_view = u.create_column(label,
                                       uuid_row,
                                       modified_row,
                                       data_source_chooser,
                                       u.create_stretch(),
                                       spacing=12)


class DisplayItemInspectorHandlerFactory(EntityBrowser.ItemInspectorHandlerFactory):
    def make_sections(self, context: EntityBrowser.Context, item: typing.Any) -> typing.Sequence[typing.Tuple[str, Declarative.HandlerLike]]:
        if isinstance(item, DisplayItem.DisplayItem):
            return [(_("Display Item"), DisplayItemInspectorHandler(context, item))]
        return list()


class DataStructureInspectorHandler(Declarative.Handler):
    def __init__(self, context: EntityBrowser.Context, data_structure: DataStructure.DataStructure):
        super().__init__()
        self.context = context
        self.data_structure = data_structure
        self.ui_view = self.__make_ui()
        self.uuid_converter = Converter.UuidToStringConverter()
        self.date_converter = Converter.DatetimeToStringConverter(is_local=True, format="%Y-%m-%d %H:%M:%S %Z")

    def __make_ui(self) -> Declarative.UIDescriptionResult:
        u = Declarative.DeclarativeUI()
        label = u.create_label(text=self.data_structure.structure_type)
        uuid_row = u.create_row(u.create_label(text="UUID:", width=60), u.create_label(text="@binding(data_structure.uuid, converter=uuid_converter)"), u.create_stretch(), spacing=12)
        modified_row = u.create_row(u.create_label(text="Modified:", width=60), u.create_label(text="@binding(data_structure.modified, converter=date_converter)"), u.create_label(text="(local)"), u.create_stretch(), spacing=12)
        entity = self.data_structure.entity
        source_line = [u.create_component_instance("source_component")] if self.context.values.get("do_references", False) else []
        if entity and entity.entity_type:
            entity_component = u.create_component_instance("entity_component")
        else:
            label2 = u.create_label(text=entity.entity_type.entity_id if entity else "NO ENTITY")
            label3 = u.create_label(text=str(entity.entity_type._field_type_map) if entity else str(self.data_structure.write_to_dict()["properties"]), max_width=300)
            entity_component = u.create_column(label2, label3, spacing=12)
        return u.create_column(label,
                               uuid_row,
                               modified_row,
                               *source_line,
                               entity_component,
                               u.create_stretch(),
                               spacing=12)

    def create_handler(self, component_id: str, container: typing.Optional[Symbolic.ComputationVariable] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "source_component":
            return EntityBrowser.ReferenceHandler(self.context, _("Source"), self.data_structure.source)
        if component_id == "entity_component":
            entity = self.data_structure.entity
            assert entity
            # note: use data structure as the base item of the entity field handler since the entity is not quite
            # properly connected to the data structure. setting a value on the data structure does not send a property
            # changed value from the entity.
            return EntityBrowser.make_record_handler(self.context, self.data_structure, entity.entity_type.entity_id, entity.entity_type)
        return None


class DataStructureInspectorHandlerFactory(EntityBrowser.ItemInspectorHandlerFactory):
    def make_sections(self, context: EntityBrowser.Context, item: typing.Any) -> typing.Sequence[typing.Tuple[str, Declarative.HandlerLike]]:
        if isinstance(item, DataStructure.DataStructure):
            return [(_("Data Structure"), DataStructureInspectorHandler(context, item))]
        return list()


class ComputationInspectorHandlerFactory(EntityBrowser.ItemInspectorHandlerFactory):
    def make_sections(self, context: EntityBrowser.Context, item: typing.Any) -> typing.Sequence[typing.Tuple[str, Declarative.HandlerLike]]:
        if isinstance(item, Symbolic.Computation):
            return [
                (_("Edit"), ComputationPanel.ComputationInspectorHandler(context, item)),
                (_("Errors"), ComputationPanel.ComputationErrorInspectorHandler(context, item)),
            ]
        return list()


Registry.register_component(DataItemInspectorHandlerFactory(), {"item-inspector-handler-factory"})
Registry.register_component(DisplayItemInspectorHandlerFactory(), {"item-inspector-handler-factory"})
Registry.register_component(DataStructureInspectorHandlerFactory(), {"item-inspector-handler-factory"})
Registry.register_component(ComputationInspectorHandlerFactory(), {"item-inspector-handler-factory"})


# since data items, display items, and other root objects are already available as top level components in the
# inspector, use an alternate schema for the project which does not include those items. this makes browsing the
# non-items in the project easier.
ProjectExtra = Schema.entity("project_extra", None, 3, {
    "title": Schema.prop(Schema.STRING),
    "workspace": Schema.reference(DataModel.Workspace),
    "data_item_references": Schema.map(Schema.STRING, Schema.reference(DataModel.DataItem)),
    "mapped_items": Schema.array(Schema.reference(DataModel.DataItem), Schema.OPTIONAL),
    "project_data_folders": Schema.array(Schema.prop(Schema.PATH)),
})


class ProjectItemsDialog(Declarative.WindowHandler):

    def __init__(self, document_controller: DocumentController.DocumentController) -> None:
        super().__init__()

        self.__document_controller = document_controller

        self.dialog_id = "project_items"

        context = Inspector.ComputationInspectorContext(document_controller, self, True)

        items = (
            EntityBrowser.EntityBrowserEntry(context, _("Data Items"), document_controller.document_model, "data_items",
                                             DataModel.DataItem, _("Data Item"), lambda x: typing.cast(str, x.title) if x.title else _("Auto")),
            EntityBrowser.EntityBrowserEntry(context, _("Display Items"), document_controller.document_model, "display_items",
                                             DataModel.DisplayItem, _("Display Item"), operator.attrgetter("displayed_title")),
            EntityBrowser.EntityBrowserEntry(context, _("Data Structures"), document_controller.document_model, "data_structures",
                                             None, _("Data Structure"), operator.attrgetter("structure_type")),
            EntityBrowser.EntityBrowserEntry(context, _("Computations"), document_controller.document_model, "computations",
                                             DataModel.Computation, _("Computation"), operator.attrgetter("label")),
            EntityBrowser.EntityBrowserEntry(context, _("Connections"), document_controller.document_model, "connections",
                                             DataModel.Connection, _("Connection"), lambda x: str(x.uuid)),
            EntityBrowser.EntityBrowserEntry(context, _("Data Groups"), document_controller.document_model, "data_groups",
                                             DataModel.DataGroup, _("Data Group"), operator.attrgetter("title")),
            EntityBrowser.EntityBrowserEntry(context, _("Projects"), typing.cast(typing.Any, document_controller.app).profile, "projects",
                                             ProjectExtra, _("Project"), lambda x: x.title or _("Project")),
            EntityBrowser.EntityBrowserEntry(context, _("Workspaces"), document_controller.project, "workspaces",
                                             DataModel.Workspace, _("Workspace"), operator.attrgetter("name")),
        )

        # close any previous list dialog associated with the window
        previous_window = getattr(document_controller, f"_{self.dialog_id}_dialog", None)
        if isinstance(previous_window, ProjectItemsDialog):
            previous_window.close_window()
        setattr(document_controller, f"_{self.dialog_id}_dialog", self)

        self.__entity_browser_component = EntityBrowser.make_entity_browser_component(items)

        u = Declarative.DeclarativeUI()

        window = u.create_window(u.create_component_instance(identifier="content"), title=_("Project Items (Computations)"), margin=12, window_style="tool")
        self.run(window, parent_window=document_controller, persistent_id=self.dialog_id)
        self.__document_controller.register_dialog(self.window)

    def close(self) -> None:
        setattr(self.__document_controller, f"_{self.dialog_id}_dialog", None)
        super().close()

    def create_handler(self, component_id: str, container: typing.Optional[Symbolic.ComputationVariable] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "content":
            return self.__entity_browser_component
        return None

    @property
    def document_model(self) -> DocumentModel.DocumentModel:
        return self.__document_controller.document_model

    def open_project_item(self, item: typing.Any) -> None:
        # in the entity browser component, focus the item, contained within `self.__document_controller.project` having type `DataModel.Project`.
        EntityBrowser.open_project_item(self.__entity_browser_component, self.__document_controller.project, DataModel.Project, "project", item)


def make_project_items_dialog(document_controller: DocumentController.DocumentController) -> None:
    ProjectItemsDialog(document_controller)


"""
Architectural Decision Records.

ADR 2023-01-10: The entity browser should be independent from the project items dialog with the expectation that it is a general
tool that can eventually be moved into the nionui framework. This implies that there is some way to supply the
entity browser with custom tabs (item-inspector-handler-factory) that are nionswift specific.

ADR 2023-01-10: The project items dialog should not be a primary tool for a typical user as it is too confusing. This implies that
the computation dialog should not provide links to open the project items dialog.

ADR 2023-01-10: The initial version of the project items dialog should be used for browsing the data structures and exploring
the links; but not be an editor. This simplifies the implementation.

ADR 2023-01-10: The project items dialog should allow a developer to browse the items and relationships in order to debug
and understand the architecture.
"""
